import copy
import json
import random
import sys
from pathlib import Path
from typing import Optional

from .enums import Brackets
from .resources import CONFIG_BASE
from .categorize import create_hero_grid


def _get_steam_userdata_path() -> Path:  
    if sys.platform == "win32":
        p = Path("C:/Program Files (x86)")
    elif sys.platform == "darwin":
        p = Path.home() / "Library/Application Support"
    elif sys.platform == "linux":
        p = Path.home()
    else:
        raise NotImplementedError("Userdata directory auto-detection is not supported for your OS!")  
    
    return p / "Steam/userdata"


def get_cfg_path(path: str) -> Path:
    try:
        cfg_path = Path(path)
    except TypeError:
        raise TypeError("User cfg directory cannot be a None value!")
    if not cfg_path.exists():
        raise ValueError(f"User cfg directory '{cfg_path}' does not exist!")
    return cfg_path


def make_herogrid(data: dict, config: dict, bracket: int) -> None:
    config_name = f"{config['config_name']} ({Brackets(bracket).name.capitalize()})"
    grouping = config["grouping"]
    sorting = config["sort"]
    path = config["path"]

    # Create a new hero grid 
    grid = create_hero_grid(data, bracket, grouping, sorting)
    grid["config_name"] = config_name
    
    update_config(grid, config_name, path)


def update_config(grid: dict, config_name: str, path: Path) -> None:
    """Updates hero grid config file in Steam userdata directory."""
    p = path/"hero_grid_config.json"

    # Load contents of file if it exists
    if p.exists():
        config = load_herogrid_config(p) # Load contents of config file if it exists
    else:
        config = None
    
    # check config contents     
    if not config or not config.get("configs"):
        config = copy.deepcopy(CONFIG_BASE) # make new config
    
    # Update existing hero grid if one exists
    for idx, c in enumerate(config["configs"]):
        if c["config_name"] == config_name:
            config["configs"][idx] = grid
            break
    else:
        config["configs"].append(grid)

    save_herogrid_config(p, config)


def load_herogrid_config(path: Path) -> dict:
    with open(path, "r") as f:
        return json.load(f)


def save_herogrid_config(path: Path, config: dict) -> None:
    with open(path, "w") as f:
        json_data = json.dumps(config, indent="\t")
        f.write(json_data)



from enum import IntEnum


class Brackets(IntEnum):
    ALL = 0
    HERALD = 1
    GUARDIAN = 2
    CRUSADER = 3
    ARCHON = 4
    LEGEND = 5
    ANCIENT = 6
    DIVINE = 7
    PRO = 8
    
    # default bracket
    DEFAULT = DIVINE


class Grouping(IntEnum):
    NONE = 0
    MAINSTAT = 1
    ATTACK = 2
    ROLE = 3

    # default grouping (Standard Dota 2 hero grid [str, int, agi])
    DEFAULT = MAINSTAT



import sys
from pathlib import Path
from typing import List, Union

import click
import yaml

from .cfg import _get_steam_userdata_path
from .enums import Brackets, Grouping
from .parseargs import parse_arg_brackets

CONF = "config.yml" # NOTE: Path("config.yml")?

CONFIG_BASE = {
    "path": None,
    "brackets": [Brackets.DEFAULT.value],
    "grouping": Grouping.DEFAULT.value,
    "config_name": "OpenDota Hero Winrates",
    "sort": "desc"
}


def _load_config(*, filename: str=None) -> dict:
    """Loads configuration file and returns it as a dict."""
    path = filename or CONF
    with open(path, "r") as f:
        config = yaml.load(f.read(), Loader=yaml.loader.FullLoader)
    if not config:
        raise ValueError
    return config


def load_config() -> dict:
    """Attempts to load configuration file. 
    If it does not exist, user is prompted to run first time setup.

    TODO: Should config["steam"]["path"] be a Path object?
    """
    try:
        config = _load_config()
    except (FileNotFoundError, ValueError):
        if click.confirm(
            "Could not find ODHG config! Do you want to run first time setup?"
            ):
            config = run_first_time_setup()
        else:
            raise SystemExit("Exiting.")
    
    # Ensure all necessary config keys are present
    config = _check_config_integrity(config)

    return config


def update_config(config: dict, *, filename: str=None) -> None:
    """Saves config as a YAML file."""
    path = filename or CONF
    with open(path, "w") as f:
        f.write(yaml.dump(config, default_flow_style=False))    


def _check_config_integrity(config: dict) -> dict:
    CONFIG_FUNCS = {
        "path": setup_userdata_cfg_dir,
        "brackets": setup_bracket,
        "grouping": setup_grouping,
        "config_name": setup_config_name,
        "sort": setup_winrate_sorting
    }

    # Remove unknown keys
    for key in config:
        if key not in CONFIG_BASE:
            config.pop(key)

    # Check for missing keys
    missing_keys = [(k, v) for (k, v) in CONFIG_BASE.items() if k not in config]
    if missing_keys:
        missing = ", ".join([key for (key, value) in missing_keys])
        click.echo(f"'config.yml' is missing the following keys: {missing}")

        # Replace missing keys in user's config
        for (key, value) in missing_keys:
            func = CONFIG_FUNCS.get(key)
            # TODO: add missing config_func handling?
            if click.confirm(
                f"Do you want add a value for the missing key '{key}'?"
                ):
                config = func(config)
            else:
                config[key] = value
                click.echo("Adding missing key with default value.")
        
        update_config(config)

    return config


def remake_config() -> None:
    """Unused"""
    p = Path(CONF)
    try:
        p.unlink()
    except FileNotFoundError:
        click.echo("Config.yml does not exist!")
        if not click.confirm("Do you want to run first time setup to create a config?"):
            return
    else:
        click.echo("Running first time setup.")       
    finally:
        config = run_first_time_setup()
        update_config(config)
        click.echo("Successfully remade config!")


def get_path_from_user() -> Path:
    """Gets a valid path from user input."""
    p = click.prompt("Path: ")
    while not Path(p).exists():
        click.echo("Path does not exist.")
        p = click.prompt("Provide an existing path or type 'q' to quit.")
        if p.lower() == "q":
            raise SystemExit("Quitting.")
    return p


def ask_steam_userdata_path() -> Path:
    if click.confirm("Would you like to specify your userdata directory manually?"):
        click.echo("\nIt should look something like this: ", nl=False)
        
        # Per-platform message
        if sys.platform == "win32":
            click.echo("'C:\\Program Files (x86)\\Steam\\userdata\\<ID>")
        elif sys.platform == "darwin":
            click.echo("~/Library/Application Support/Steam/userdata/<ID>")
        elif sys.platform == "linux":
            click.echo("~/Steam/userdata/<ID>")
        
        return get_path_from_user()
    else:
        raise SystemExit("Unable to locate Steam userdata directory.") # abort


def setup_userdata_cfg_dir(config: dict) -> dict:
    """Configure user's Dota userdata cfg directory.
    """
    # Get default directory for Steam userdata
    click.echo("\nSteam Userdata Directory:")
    try:
        p = _get_steam_userdata_path()
    except NotImplementedError as e:
        click.echo(e.args[0])
        cfg_path = ask_steam_userdata_path()
    else:
        directories = [d for d in p.iterdir()]
        
        # Ask user for a path if no directories found in auto-detected path
        if not directories:
            click.echo(f"No subdirectories were found in {p}")
            cfg_path = ask_steam_userdata_path()
        
        else:
            # Subdirectories
            subdirs = {idx+1: d for idx, d in enumerate(directories)}
            
            # Let user select a directory
            _choices = "\n".join(f"\t{idx}. {d.stem}" for idx, d in subdirs.items())
            click.echo(_choices)
            
            choice = 0
            while choice not in subdirs:
                choice = click.prompt(f"Select directory (1-{len(directories)})", type=int) 
            cfg_path = subdirs.get(choice)
    finally:    
        config["path"] = str((cfg_path / "570/remote/cfg"))
    
    return config


def setup_bracket(config: dict) -> dict:
    # Determine lowest and highest Bracket enum value
    _b_values = [b.value for b in Brackets]
    brackets_start = min(_b_values)
    brackets_end = max(_b_values)
    
    # Prompt user to select a default bracket
    available_brackets = get_enum_string(Brackets)
    click.echo(f"\nBrackets:\n{available_brackets}")

    brackets = _get_brackets(
        "Specify default skill bracket(s), separated by spaces", 
        brackets_start, 
        brackets_end
    )
    while not brackets:
        brackets = _get_brackets(
            "No valid brackets provided. Try again", 
            brackets_start, 
            brackets_end
        )

    config["brackets"] = brackets

    return config


def _get_brackets(msg: str, brackets_start: int, brackets_end: int) -> List[int]:
    b = click.prompt(
        f"{msg} ({brackets_start}-{brackets_end})",
        type=str,
        default=str(Brackets.DEFAULT.value),
        show_default=False
    )
    return parse_arg_brackets(b.split(" "))


def setup_grouping(config: dict) -> dict:
    # Determine lowest and highest Grouping enum value
    _g_values = [g.value for g in Grouping]
    grouping_start = min(_g_values)
    grouping_end = max(_g_values)
    
    # Prompt user to select a default grouping
    click.echo(get_enum_string(Grouping))

    
    get_grp = lambda m: click.prompt(
        f"{m} ({grouping_start}-{grouping_end})",
        type=int,
        default=Grouping.DEFAULT.value,
        show_default=False
    )  
    grouping = get_grp("\nSelect default hero grouping") 
    while grouping not in _g_values:
        grouping = get_grp("Invalid grouping. Try again")
    
    config["grouping"] = grouping

    return config


def get_enum_string(enum: Union[Brackets, Grouping]) -> str:
    choices = "\n".join(
        f"\t{e.value}. {e.name.capitalize()}" 
        if e != enum.DEFAULT else 
        f"\t{e.value}. {e.name.capitalize()} [default]" 
        for idx, e in enumerate(enum)
    )
    return choices


def setup_config_name(config: dict) -> dict:
    """Setup for default name of hero grid."""
    click.echo(
        f"\nChoose a default hero grid name. (current: {config['config_name']})"
        )
    name = click.prompt(
        f"New name (leave blank to keep current name)", 
        default="",
        show_default=False
        )
    if name:
        config["config_name"] = name
    return config


def setup_winrate_sorting(config: dict) -> dict:
    click.echo("\nDo you want to sort heroes by winrates descending or ascending?")
    click.echo("1. Descending [default]")
    click.echo("2. Ascending")
    
    choice = 0
    while choice not in [1, 2]:
        choice = click.prompt("Choice (1-2)", type=int, default=1, show_default=False)
    config["sort"] = (choice == 1)

    return config


def run_first_time_setup() -> dict:
    # Create new config file
    p = Path(CONF)
    if Path(CONF).exists():
        if not click.confirm(
            f"'{p}' already exists. Are you sure you want to overwrite it?"
        ):
            click.echo("Aborting setup.")
            raise SystemExit
    
    click.echo("Creating new config...")  
    config = CONFIG_BASE

    # Setup config parameters
    config = setup_userdata_cfg_dir(config)
    config = setup_bracket(config)
    config = setup_grouping(config)
    config = setup_winrate_sorting(config)
    config = setup_config_name(config)

    update_config(config)

    return config



from pathlib import Path

import click

from .cfg import make_herogrid, get_cfg_path
from .config import load_config, run_first_time_setup
from .odapi import fetch_hero_stats
from .parseargs import parse_arg_brackets, parse_arg_grouping


def get_help_string() -> dict:
    """Gets help string from README.md"""
    p = Path(__file__)
    with open(p.parent.parent / "README.md", "r") as f:
        readme = f.read()
    lines = []
    active = False
    start_md = False
    for line in readme.splitlines():
        if line.startswith("# Usage"):
            active = True
            continue
        if active:
            if line.startswith("```"):
                if not start_md:
                    start_md = True
                    continue
                else:
                    break # break out once we have reached end of codeblock
            else:
                lines.append(line)
    return "\n".join(lines)



def get_config_from_cli_arguments(**options) -> dict:
    """Fills missing arguments using values from 'config.yml'.
    
    Returns config
    """
    
    if all(v for v in options.values()):
        config = options
    else: # Load config.yml if not all CLI arguments are passed in
        config = load_config()
        # Fill config with CLI arguments
        for option, argument in options.items():
            if argument or argument == 0: # we need to accept 0
                config[option] = argument

    # Parse config values
    config = parse_config(config)
    
    return config


def parse_config(config: dict) -> dict:
    config["brackets"] = parse_arg_brackets(config["brackets"])
    config["grouping"] = parse_arg_grouping(config["grouping"])
    
    # We can fall back on bracket and grouping defaults
    # But we can't fall back on a default Steam userdata directory path
    try:
        config["path"] = get_cfg_path(config["path"]) # Steam userdata directory
    except (TypeError, ValueError) as e:
        click.echo(e.args[0])
        click.echo(
            "Either specify a valid path using the '--path' option, "
            "or run setup using the '--setup' flag to permanently add "
            "a valid path to your config."
        )
        raise SystemExit
    return config


@click.command()
@click.option("--brackets", "-b", default=None, multiple=True)
@click.option("--grouping", "-g", default=None)
@click.option("--path", "-p", default=None) # we forego click.Path here and do our own check
@click.option("--sort", "-s", is_flag=True, default=True) # enable for ascending sorting # TODO: rename sort to one of [sort, sortasc, asc, ascending]
@click.option("--setup", "-S", is_flag=True)
@click.option("--help", is_flag=True)
def main(**options) -> None:
    if options.pop("help"):
        click.echo("USAGE:")
        click.echo(get_help_string())
        raise SystemExit
    
    if options.pop("setup", None):
        options = run_first_time_setup()
        # Ask if user wants to generate grid?
    
    config = get_config_from_cli_arguments(**options)

    # Fetch hero W/L stats from API
    data = fetch_hero_stats()
    
    # Create grid for each specified bracket
    for bracket in config["brackets"]:
        make_herogrid(data, config, bracket)
    

# TODO: Config class?

if __name__ == "__main__":   
    main()



import copy
from typing import Union

from .enums import Grouping
from .resources import CONFIG, CATEGORY


def create_hero_grid(heroes: list, bracket: int, grouping: int, sorting: bool) -> dict:
    grouping_funcs = {
        Grouping.MAINSTAT.value: group_by_main_stat,
        Grouping.NONE.value: group_by_all,
        Grouping.ATTACK.value: group_by_melee_ranged,
        Grouping.ROLE.value: group_by_role
    }
    # Sort heroes by winrate in the specified bracket
    heroes = sort_heroes_by_winrate(heroes, bracket, sorting)       
    
    grp_func = grouping_funcs.get(grouping)
    if not grp_func:
        raise ValueError(f"No such grouping: '{grouping}'")
    
    config = grp_func(heroes)
    
    return config


def sort_heroes_by_winrate(heroes: list, bracket: str, descending: bool=True) -> list:
    """Sorts list of heroes by winrate in a specific skill bracket."""
    for hero in heroes:
        if hero["8_pick"] == 0:
            hero["8_pick"] = 1
    heroes.sort(key=lambda h: h[f"{bracket}_win"] / h[f"{bracket}_pick"], reverse=descending)
    return heroes


def group_by_main_stat(heroes: list) -> dict:
    """Get hero grid, categorized by main stat."""
    CATEGORY_IDX = {
        "str": 0,
        "agi": 1,
        "int": 2
    }
    config = _get_new_config()
    for hero in heroes:
        idx = CATEGORY_IDX.get(hero["primary_attr"])
        config["categories"][idx]["hero_ids"].append(hero["id"])
    return config


def group_by_all(heroes: list) -> dict:
    """Get hero grid, all heroes together in a single category."""
    category = _get_new_category("Heroes", height=1180.0)

    config = _get_new_config()
    config["categories"] = [category] # Override predefined categories

    for hero in heroes:
        config["categories"][0]["hero_ids"].append(hero["id"])
    return config


def group_by_melee_ranged(heroes: list) -> dict:
    """Get hero grid, categorized by main melee/ranged."""
    melee = _get_new_category("Melee", height=280.0)
    ranged = _get_new_category("Ranged", y_pos=300.0, height=280.0)

    config = _get_new_config()
    config["categories"] = [melee, ranged] # Override predefined categories

    for hero in heroes:
        idx = 1 if hero["attack_type"] == "Ranged" else 0
        config["categories"][idx]["hero_ids"].append(hero["id"])
    return config


def group_by_role(heroes: list) -> dict:
    """Get hero grid, categorized by carry/support/flex."""
    carry = _get_new_category("Carry")
    support = _get_new_category("Support", y_pos=200.0)
    flex = _get_new_category("Flexible", y_pos=400.0)

    config = _get_new_config()
    config["categories"] = [carry, support, flex] # Override predefined categories

    for hero in heroes:
        if "Carry" in hero["roles"]:
            idx = 0 
        elif "Support" in hero["roles"]:
            idx = 1
        else:
            idx = 2
        config["categories"][idx]["hero_ids"].append(hero["id"])
    return config


def _get_new_category(name: str, x_pos: float=0.0, y_pos: float=0.0, width: float=0.0, height: float=0.0) -> dict:
    # Copy category and give it a name
    category = copy.deepcopy(CATEGORY)
    category["category_name"] = name

    params = {
        "x_position": x_pos,
        "y_position": y_pos,
        "width": width,
        "height": height
    }
    for param, value in params.items():
        value = float(abs(value)) # ensure value is a positive float
        if value:
            category[param] = value

    return category


def _get_new_config() -> dict:
    return copy.deepcopy(CONFIG)



"""Not to be confused with the argparse module in the Python standard library.
You can easily tell the difference, because this module is a piece of shit.
"""

from enum import IntEnum
from typing import List, Optional, Union, Dict, Tuple

import click

from .enums import Brackets, Grouping


def make_mapping(mapping: Dict[Tuple[str], IntEnum]) -> Dict[Union[str, int], IntEnum]:
    """Make a new mapping with multiple keys for identical enums."""
    new = {}
    for keys, enum in mapping.items():
        new[enum.value] = enum.value         # add Enum value as key
        new[enum.name.lower()] = enum.value  # add Enum name as key
        for key in keys:
            new[key] = enum.value            # add existing keys from mapping
    return new


_brackets = {
    ("h",): Brackets.HERALD,
    ("g",): Brackets.GUARDIAN,
    ("c",): Brackets.CRUSADER,
    ("a",): Brackets.ARCHON,
    ("l",): Brackets.LEGEND,
    ("n",): Brackets.ANCIENT,
    ("d", "immortal", "i"): Brackets.DIVINE,
    ("p", "official", "tournaments"): Brackets.PRO,
    ("A", "all"): Brackets.ALL,
}

_grouping = {
    ("m", "stat", "mainstat", "stats"): Grouping.MAINSTAT,
    ("a", "melee", "range", "attack"): Grouping.ATTACK,
    ("r",): Grouping.ROLE,
    ("n", "all", "everything"): Grouping.NONE,
}


GROUPING = make_mapping(_grouping)
BRACKETS = make_mapping(_brackets)


def parse_arg_brackets(brackets: List[Union[str, int]]) -> List[int]:
    """Parses bracket (`-b` `--bracket`) argument.
    
    Returns list of integers.
    """
    # FIXME: this is a mess now

    # Check is Brackets.ALL is given as an argument
    if any(x in brackets for x in [str(Brackets.ALL.value), Brackets.ALL.value]):
        # Create list with all brackets
        valid_brackets = [b.value for b in Brackets if b.value != Brackets.ALL.value]
    else:
        # Parse arguments & make list of valid bracket arguments
        valid_brackets = list(
            filter(
                None.__ne__, [find_argument_in_mapping(b, BRACKETS) for b in brackets]
            )
        )

    # Fall back on default value if no valid brackets are provided
    if not valid_brackets:
        valid_brackets = [Brackets.DEFAULT.value]
        click.echo(
            "No valid bracket arguments provided. "
            f"Using default bracket: {Brackets.DEFAULT.name.capitalize()}"
        )

    return list(set(valid_brackets))


def parse_arg_grouping(grouping: str) -> int:
    """Parses grouping (`-g` `--group`) argument.
    
    Returns integer
    """
    grp = find_argument_in_mapping(grouping, GROUPING)
    if grp is None:
        grp = Grouping.DEFAULT.value
        click.echo(
            "No valid grouping arguments provided. "
            f"Using default bracket: {Grouping.DEFAULT.name.capitalize()}"
        )
    return grp


def find_argument_in_mapping(argument: Union[str, int], mapping: dict) -> Optional[int]:
    if isinstance(argument, str) and argument.isdigit():
        argument = int(argument)
    return mapping.get(argument)  # None is default




DEFAULT_NAME = "OpenDota Hero Winrates"

DOTA_GRID = {
    "version": 3,
    "configs": [
        {
            "config_name": DEFAULT_NAME,
            "categories": [
                {
                    "category_name": "Strength",
                    "x_position": 0.000000,
                    "y_position": 0.000000,
                    "width": 1180.0,
                    "height": 180.0,
                    "hero_ids": [],
                },
                {
                    "category_name": "Agility",
                    "x_position": 0.000000,
                    "y_position": 200.000000,
                    "width": 1180.0,
                    "height": 180.0,
                    "hero_ids": [],
                },
                {
                    "category_name": "Intelligence",
                    "x_position": 0.000000,
                    "y_position": 400.000000,
                    "width": 1180.0,
                    "height": 180.0,
                    "hero_ids": [],
                }
            ],
        }
    ],
}

CATEGORY = {
    "category_name": "name",
    "x_position": 0.000000,
    "y_position": 0.000000,
    "width": 1180.0,
    "height": 180.0,
    "hero_ids": [],
}

CONFIG = {
    "config_name": DEFAULT_NAME,
    "categories": [
        {
            "category_name": "Strength",
            "x_position": 0.000000,
            "y_position": 0.000000,
            "width": 1180.0,
            "height": 180.0,
            "hero_ids": [],
        },
        {
            "category_name": "Agility",
            "x_position": 0.000000,
            "y_position": 200.000000,
            "width": 1180.0,
            "height": 180.0,
            "hero_ids": [],
        },
        {
            "category_name": "Intelligence",
            "x_position": 0.000000,
            "y_position": 400.000000,
            "width": 1180.0,
            "height": 180.0,
            "hero_ids": [],
        }
    ]
}

CONFIG_TWO_CATEGORIES = {
    "config_name": DEFAULT_NAME,
    "categories": [
        {
            "category_name": "Melee",
            "x_position": 0.000000,
            "y_position": 0.000000,
            "width": 1180.0,
            "height": 280.0,
            "hero_ids": [],
        },
        {
            "category_name": "Ranged",
            "x_position": 0.000000,
            "y_position": 300.000000,
            "width": 1180.0,
            "height": 280.0,
            "hero_ids": [],
        },
    ]
}


CONFIG_THREE_CATEGORIES = {
    "config_name": DEFAULT_NAME,
    "categories": [
        {
            "category_name": "Strength",
            "x_position": 0.000000,
            "y_position": 0.000000,
            "width": 1180.0,
            "height": 180.0,
            "hero_ids": [],
        },
        {
            "category_name": "Agility",
            "x_position": 0.000000,
            "y_position": 200.000000,
            "width": 1180.0,
            "height": 180.0,
            "hero_ids": [],
        },
        {
            "category_name": "Intelligence",
            "x_position": 0.000000,
            "y_position": 400.000000,
            "width": 1180.0,
            "height": 180.0,
            "hero_ids": [],
        }
    ]
}


CONFIG_BASE = {
    "version": 3,
    "configs": [],
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['odherogrid']

package_data = \
{'': ['*'], 'odherogrid': ['scripts/*']}

install_requires = \
['click>=7.0,<8.0', 'pyyaml>=5.2,<6.0', 'requests>=2.22.0,<3.0.0']

entry_points = \
{'console_scripts': ['odhg = odherogrid.scripts.cli:run']}

setup_kwargs = {
    'name': 'odherogrid',
    'version': '0.1.0',
    'description': 'Dota 2 hero grid generator using OpenDotaAPI stats',
    'long_description': None,
    'author': 'PederHA',
    'author_email': 'peder.andresen@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/PederHA/odherogrid',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import requests

from .enums import Brackets

def fetch_hero_stats() -> list:
    """Retrieves hero win/loss statistics from OpenDotaAPI."""
    r = requests.get("https://api.opendota.com/api/heroStats")
    heroes = r.json()
    # Rename pro_<stat> to 8_<stat>, so it's easier to work with our enum
    for hero in heroes:
        for stat in ["win", "pick", "ban"]:
            hero[f"{Brackets.PRO.value}_{stat}"] = hero.pop(f"pro_{stat}")
    return heroes


__version__ = '0.1.0'

from .categorize import *
from .cfg import *
from .config import *
from .enums import *
from .odapi import *
from .odhg import *
from .parseargs import *
from .resources import *